from enum import Enum


class Content_type:
    text = "text"
    user_phone_number = "user_phone_number"
    user_email = "user_email"


class Type:
    postback = "postback"
    web_url = "web_url"
    phone_number = "phone_number"
    account_link = "account_link"
    account_unlink = "account_unlink"


class Notification_frequency(Enum):
    """
    Message frequency for the subscription message.
    """

    DAILY = "DAILY"
    WEEKLY = "WEEKLY"
    MONTHLY = "MONTHLY"


class Notification_cta_text(Enum):
    """
    Call to action text for the subscription message.
    """

    ALLOW = "ALLOW"
    FREQUENCY = "FREQUENCY"
    GET = "GET"
    OPT_IN = "OPT_IN"
    SIGN_UP = "SIGN_UP"


class Notification_reoptin(Enum):
    """
    Reoptin action for the subscription message.
    """

    ENABLED = "ENABLED"
    DISABLED = "DISABLED"


class Action:
    mark_seen = "mark_seen"
    typing_on = "typing_on"
    typing_off = "typing_off"


class Filetype:
    file = "file"
    video = "video"
    image = "image"
    audio = "audio"


class Messaging_type:
    MESSAGE_TAG = "MESSAGE_TAG"
    RESPONSE = "RESPONSE"
    UPDATE = "UPDATE"


class Tag:
    ACCOUNT_UPDATE = "ACCOUNT_UPDATE"
    CONFIRMED_EVENT_UPDATE = "CONFIRMED_EVENT_UPDATE"
    CUSTOMER_FEEDBACK = "CUSTOMER_FEEDBACK"
    HUMAN_AGENT = "HUMAN_AGENT"
    POST_PURCHASE_UPDATE = "POST_PURCHASE_UPDATE"


class Notification_type:
    NO_PUSH = "NO_PUSH"
    REGULAR = "REGULAR"
    SILENT_PUSH = "SILENT_PUSH"



import codecs
import pickle
import urllib.parse
from .cmd import Cmd


class Payload:
    """
    Object for Payload Management
    """

    def __init__(self, payload, **kwargs) -> None:
        """
        Object for Payload Management
        """
        self.payload = payload
        self.data = kwargs

    def __str__(self):
        return Payload.trt_payload_out(self)

    @staticmethod
    def trt_payload_in(payload0):
        """
        processing of payloads received in a sequence of structured parameters

        @params: payload [String]
        @return: payload [String] , structured parameters Dict
        """

        payload = urllib.parse.unquote(payload0)

        res = {}
        while "{{" in payload:
            start = payload.index("{{")
            end = payload.index("}}")
            items = payload[start + 2 : end].split("===")
            # result string to object
            res[items[0]] = pickle.loads(codecs.decode(items[1].encode(), "base64"))
            payload = payload.replace(payload[start : end + 2], "").strip()
        return (
            payload0.copy(payload) if isinstance(payload0, Cmd) else payload,
            res,
        )

    @staticmethod
    def trt_payload_out(payload):
        """
        Processing of a Payload type as a character string

        @params: payload [ Payload | String ]
        @return: String
        """
        if isinstance(payload, Payload):
            tmp = ""
            for key_data, val_data in payload.data.items():
                # object to string
                val_data = codecs.encode(pickle.dumps(val_data), "base64").decode()
                tmp += f"{{{{{key_data}==={val_data}}}}} "

            final_pl = payload.payload + (" " + tmp if tmp else "")
            if len(final_pl) >= 2000:
                raise Exception("Payload data is too large")
            return urllib.parse.quote(final_pl)
        return urllib.parse.quote(payload)



class Cmd(str):
    """
    Object for text of message
    """

    webhook = "message"
    token = None
    __atts = []

    def __init__(self, text):
        str.__init__(text)

    def set_atts(self, atts):
        for att in atts:
            self.__atts.append(att)

    @property
    def attachments(self):
        return self.__atts

    def copy(self, text):
        new_cmd = Cmd(text)
        new_cmd.__atts = self.attachments
        new_cmd.webhook = self.webhook
        new_cmd.token = self.token
        return new_cmd



import os
import sys
import json
import requests
from .payload import Payload
from conf import Configuration  # type: ignore


LANGS = None


def download_file(url, file):
    """
    Downloading a file from an url.

    Args:
        url: direct link for the attachment

        file: filename with path
    """
    res = requests.get(url, allow_redirects=True)

    with open(file, "wb") as f:
        f.write(res.content)

    return file


def translate(key, lang):
    """
    this function uses the langs.json file.
    translate a keyword or sentence
    """
    global LANGS

    if not lang:
        return key

    if not os.path.isfile("langs.json"):
        print("Warning! langs.json not found", file=sys.stderr)
        from .source import langs

        with open("langs.json", "w", encoding="utf-8") as lang_file:
            lang_file.write(langs)
            print("langs.json created!")
        return key

    if not LANGS:
        with open("langs.json", encoding="utf-8") as lang_file:
            LANGS = json.load(lang_file)

    keyword = LANGS.get(key)

    if keyword:
        if keyword.get(lang):
            return keyword.get(lang)
    return key


def simulate(sender_id, text, **params):
    """
    Simulate a message send by an user

    Args:
        sender_id: <str>
        text: <str> |<Payload>
    """
    if isinstance(text, Payload):
        text = Payload.trt_payload_out(text)

    data_json = {
        "object": "page",
        "entry": [
            {
                "messaging": [
                    {
                        "message": {
                            "text": text,
                        },
                        "sender": {"id": sender_id},
                    }
                ]
            }
        ],
    }
    header = {"content-type": "application/json; charset=utf-8"}
    return requests.post(
        f"http://127.0.0.1:{Configuration.APP_PORT}/",
        json=data_json,
        headers=header,
        params=params,
    )



import os
import json
import pickle
import requests
from retry import retry
import requests_toolbelt
from .logger import Logger
from .payload import Payload
from conf import Configuration  # type: ignore
from .ui import ReceiptElement, Summary, Address, Adjustment
from .constant import Tag, Action, Filetype, Messaging_type, Notification_type
from .ui import (
    Button,
    QuickReply,
    Element,
    Product,
    RecurringNotificationOptin,
)


class Messenger:
    def __init__(self, log_level="error", api_version="v15.0"):
        """
        Args:
            log_level (error, info, quiet): A type of display log
                error: print only error
                info: print access log and error log
                quiet: do not print anything
        """
        self.access_token = Configuration.ACCESS_TOKEN
        self.api_version = api_version
        self.url = f"https://graph.facebook.com/{api_version}/me"

        if log_level not in ("error", "info", "quiet"):
            raise Exception(ValueError, "log_level must be error or info or quiet")

    def __analyse(self, res, log_level="error"):
        if log_level == "info":
            Logger.info(f"\n  status_code : {res.status_code}, data :{res.text}")
        elif res.status_code != 200 and log_level == "error":
            Logger.error(f"\n  status_code : {res.status_code}, data :{res.text}")
        return res

    @property
    def token(self):
        if not self.access_token:
            Logger.warning("No access token provided")
        return self.access_token

    @property
    def page_id(self):
        """
        This method is used to get the page id of the facebook page
        """
        res = requests.get(f"{self.url}?access_token={self.token}&fields=id")
        return self.__analyse(res).json()["id"]

    @retry(requests.exceptions.ConnectionError, tries=3, delay=3)
    def send_custom(self, json, endpoint="/messages"):
        """
        This method send a custom json,
        use for the api not implemented in ampalibe.

        Args:
            json: A json object
            endpoint: A endpoint of facebook api
                /messages by default like other endpoint
        """
        header = {"content-type": "application/json; charset=utf-8"}
        params = {"access_token": self.token}

        res = requests.post(
            self.url + endpoint, json=json, headers=header, params=params
        )
        return self.__analyse(res)

    @retry(requests.exceptions.ConnectionError, tries=3, delay=3)
    def send_message(self, sender_id, message, prio=False):
        Logger.warning("This method is deprecated, use send_text instead")
        return self.send_text(sender_id, message)

    @retry(requests.exceptions.ConnectionError, tries=3, delay=3)
    def send_text(self, sender_id, text, **kwargs):
        """
        This method allows you to send a text message to the given recipient,
        Note that the number of characters to send is limited to 2000 characters

        Args:
            sender_id (str): user id facebook for the destination
            text (str): message want to send

        Returns:
            Response: POST request to the facebook API to send a message to the user

        Ref:
            https://developers.facebook.com/docs/messenger-platform/send-messages#sending_text
        """

        recipient = (
            {"one_time_notif_token": kwargs.get("one_time_notif_token")}
            if kwargs.get("one_time_notif_token")
            else {"id": sender_id}
        )

        dataJSON = {"recipient": recipient, "message": {"text": text}}
        dataJSON.update(kwargs)

        header = {"content-type": "application/json; charset=utf-8"}
        params = {"access_token": self.token}

        res = requests.post(
            self.url + "/messages",
            json=dataJSON,
            headers=header,
            params=params,
        )
        return self.__analyse(res)

    @retry(requests.exceptions.ConnectionError, tries=3, delay=3)
    def send_attachment(self, sender_id, attachment_id, filetype="file", **kwargs):
        """
        The Messenger Platform supports saving assets via the Send API and Attachment Upload API. This allows you reuse assets, rather than uploading them every time they are needed.
        To attach a saved asset to a message, specify the attachment_id of the asset in the payload.attachment_id property of the message request:

        Args:
            sender_id (str): user id facebook for the destination
            attachment_id (str): The reusable attachment ID
            filetype (str, optional): type of the file["video","image",...]. Defaults to 'file'.

        Returns:
            Response: POST request to the facebook API to send a message to the user

        Ref:
            https://developers.facebook.com/docs/messenger-platform/reference/attachment-upload-api#attachment_reuse
        """
        recipient = (
            {"one_time_notif_token": kwargs.get("one_time_notif_token")}
            if kwargs.get("one_time_notif_token")
            else {"id": sender_id}
        )

        dataJSON = {
            "recipient": recipient,
            "message": {
                "attachment": {
                    "type": filetype,
                    "payload": {"attachment_id": attachment_id},
                }
            },
        }
        dataJSON.update(kwargs)

        header = {"content-type": "application/json; charset=utf-8"}
        params = {"access_token": self.token}

        res = requests.post(
            self.url + "/messages",
            json=dataJSON,
            headers=header,
            params=params,
        )
        return self.__analyse(res)

    @retry(requests.exceptions.ConnectionError, tries=3, delay=3)
    def send_action(self, sender_id, action, **kwargs):
        """
        This method is used to simulate an action on messages.
        example: view, writing.
        Action available: ['mark_seen', 'typing_on', 'typing_off']

        Args:
            sender_id (str): user id facebook for the destination
            action (str): action ['mark_seen', 'typing_on', 'typing_off']

        Returns:
            Response: POST request to the facebook API to send an action to the user

        Ref:
            https://developers.facebook.com/docs/messenger-platform/send-messages/sender-actions
        """
        recipient = (
            {"one_time_notif_token": kwargs.get("one_time_notif_token")}
            if kwargs.get("one_time_notif_token")
            else {"id": sender_id}
        )

        dataJSON = {
            "messaging_type": "RESPONSE",
            "recipient": recipient,
            "sender_action": action,
        }
        dataJSON.update(kwargs)

        header = {"content-type": "application/json; charset=utf-8"}
        params = {"access_token": self.token}

        res = requests.post(
            self.url + "/messages",
            json=dataJSON,
            headers=header,
            params=params,
        )
        return self.__analyse(res)

    @retry(requests.exceptions.ConnectionError, tries=3, delay=3)
    def send_quick_reply(self, sender_id, quick_reps, text, next=None, **kwargs):
        """
        Quick replies provide a way to present a set of up to 13 buttons
        in-conversation that contain a title and optional image, and appear
        prominently above the composer. You can also use quick replies
        to request a person's location, email address, and phone number.

        Args:
            sender_id (str): user id facebook for the destination
            quick_rep (list of dict) || (list of QuickRep): list of the different quick_reply to send a user
            text (str): A text of a little description for each <quick_reply>
            next(bool) || (text): this params activate the next page when elements have a length more than ten

        Returns:
            Response: POST request to the facebook API to send a quick_reply to the user

        Ref:
            https://developers.facebook.com/docs/messenger-platform/send-messages/quick-replies
        """
        recipient = (
            {"one_time_notif_token": kwargs.get("one_time_notif_token")}
            if kwargs.get("one_time_notif_token")
            else {"id": sender_id}
        )

        quick_reps = [
            quick_rep.value if isinstance(quick_rep, QuickReply) else quick_rep
            for quick_rep in quick_reps
        ]

        dataJSON = {
            "messaging_type": "RESPONSE",
            "recipient": recipient,
            "message": {"text": text, "quick_replies": quick_reps[:13]},
        }
        dataJSON.update(kwargs)

        if len(quick_reps) > 13 and next:
            dataJSON["message"]["quick_replies"][12] = {
                "content_type": "text",
                "title": "More" if next == True else str(next),
                "payload": "/__more",
                "image_url": "https://icon-icons.com/downloadimage.php"
                + "?id=81300&root=1149/PNG/512/&file="
                + "1486504364-chapter-controls-forward-play"
                + "-music-player-video-player-next_81300.png",
            }

            pickle.dump(
                (quick_reps[12:], text, next),
                open(f"assets/private/.__{sender_id}", "wb"),
            )

        header = {"content-type": "application/json; charset=utf-8"}
        params = {"access_token": self.token}

        res = requests.post(
            self.url + "/messages",
            json=dataJSON,
            headers=header,
            params=params,
        )
        return self.__analyse(res)

    @retry(requests.exceptions.ConnectionError, tries=3, delay=3)
    def send_template(self, sender_id, elements, quick_rep=None, next=None, **kwargs):
        """
        This method is used to send a template to the user
        """
        Logger.warning("This method is deprecated, use send_generic_template instead")
        return self.send_generic_template(
            sender_id, elements, quick_rep, next, **kwargs
        )

    @retry(requests.exceptions.ConnectionError, tries=3, delay=3)
    def send_generic_template(
        self, sender_id, elements, quick_rep=None, next=None, **kwargs
    ):
        """
        The method represent a Message templates who offer a way for you
        to offer a richer in-conversation experience than standard text messages by integrating
        buttons, images, lists, and more alongside text a single message. Templates can be use for
        many purposes, such as displaying product information, asking the messagerecipient to choose
        from a pre-determined set of options, and showing search results.

        For this, messenger only validates 10 templates
        for the first display, so we put the parameter
        <next> to manage these numbers if it is a number of
        elements more than 10,

        So, there is a quick_reply which acts as a "next page"
        displaying all requested templates

        Args:
            sender_id (str): user id facebook for the destination
            elements (list of dict) || (list of Element): the list of the specific elements to define the
            structure for the template
            quick_rep (list of dict) || (list of QuickReply): addition quick reply at the bottom of the template
            next(bool) || (text): this params activate the next page when elements have a length more than ten

        Returns:
            Response: POST request to the facebook API to send a template generic to the user

        Ref:
            https://developers.facebook.com/docs/messenger-platform/send-messages/template/generic
        """
        recipient = (
            {"one_time_notif_token": kwargs.get("one_time_notif_token")}
            if kwargs.get("one_time_notif_token")
            else {"id": sender_id}
        )

        elements = [
            element.value if isinstance(element, Element) else element
            for element in elements
        ]

        dataJSON = {
            "messaging_type": "RESPONSE",
            "recipient": recipient,
            "message": {
                "attachment": {
                    "type": "template",
                    "payload": {
                        "template_type": "generic",
                        "elements": elements[:10],
                    },
                },
            },
        }
        dataJSON.update(kwargs)

        if len(elements) > 10 and next:
            dataJSON["message"]["quick_replies"] = [
                {
                    "content_type": "text",
                    "title": "Next" if next == True else str(next),
                    "payload": "/__next",
                    "image_url": "https://icon-icons.com/downloadimage.php"
                    + "?id=81300&root=1149/PNG/512/&file="
                    + "1486504364-chapter-controls-forward-play"
                    + "-music-player-video-player-next_81300.png",
                }
            ]
            pickle.dump(
                (elements[10:], next),
                open(f"assets/private/.__{sender_id}", "wb"),
            )

        if quick_rep:
            quick_rep = [
                qr.value if isinstance(qr, QuickReply) else qr for qr in quick_rep
            ]

            if isinstance(dataJSON["message"].get("quick_replies"), list):
                dataJSON["message"]["quick_replies"] = (
                    quick_rep + dataJSON["message"]["quick_replies"]
                )
            else:
                dataJSON["message"]["quick_replies"] = quick_rep

        header = {"content-type": "application/json; charset=utf-8"}
        params = {"access_token": self.token}

        res = requests.post(
            self.url + "/messages",
            json=dataJSON,
            headers=header,
            params=params,
        )
        return self.__analyse(res)

    @retry(requests.exceptions.ConnectionError, tries=3, delay=3)
    def send_file_url(self, sender_id, url, filetype="file", reusable=False, **kwargs):
        """
        The Messenger Platform allows you to attach assets to messages, including audio,
        video, images, and files.All this is the role of this Method. The maximum attachment
        size is 25 MB. The maximum resolution for images is 85 Megapixel. There are three ways
        to attach an asset to a message:

        Args:
            sender_id (str): user id facebook for destination
            url (str): the origin url for the file
            filetype (str, optional): type of the file["video","image",...]. Defaults to 'file'.
            reusable (bool, default False): Make an attachment reusable

        Returns:
            Response: POST request to the facebook API to send a template generic to the user

        Ref:
            https://developers.facebook.com/docs/messenger-platform/send-messages#url
        """
        recipient = (
            {"one_time_notif_token": kwargs.get("one_time_notif_token")}
            if kwargs.get("one_time_notif_token")
            else {"id": sender_id}
        )

        dataJSON = {
            "messaging_type": "RESPONSE",
            "recipient": recipient,
            "message": {
                "attachment": {
                    "type": filetype,
                    "payload": {"url": url, "is_reusable": reusable},
                }
            },
        }
        dataJSON.update(kwargs)
        header = {"content-type": "application/json; charset=utf-8"}
        params = {"access_token": self.token}
        res = requests.post(
            self.url + "/messages",
            json=dataJSON,
            headers=header,
            params=params,
        )
        return self.__analyse(res)

    @retry(requests.exceptions.ConnectionError, tries=3, delay=3)
    def persistent_menu(self, sender_id, menu, action="PUT", **kwargs):
        """
        The Persistent Menu disabling the composer best practices allows you to have an always-on
        user interface element inside Messenger conversations. This is an easy way to help people
        discover and access the core functionality of your Messenger bot at any point in the conversation

        Args:
            sender_id (str): user id for destination
            persistent_menu (list of dict) || (list of Button): the elements of the persistent menu to enable
            action (str, optional): the action for benefit["PUT","DELETE"]. Defaults to 'PUT'.

            locale [optionnel]
            composer_input_disabled [optionnel]

        Ref:
            https://developers.facebook.com/docs/messenger-platform/send-messages/persistent-menu
        """
        header = {"content-type": "application/json; charset=utf-8"}
        params = {"access_token": self.token}

        menu = [
            button.value if isinstance(button, Button) else button for button in menu
        ]

        if action == "DELETE":
            params["params"] = "(persistent_menu)"
            params["psid"] = sender_id

            res = requests.delete(
                self.url + "/custom_user_settings",
                headers=header,
                params=params,
            )
            return self.__analyse(res)
        else:
            dataJSON = {
                "psid": sender_id,
                "persistent_menu": [
                    {
                        "locale": kwargs.get("locale", "default"),
                        "composer_input_disabled": kwargs.get(
                            "composer_input_disabled", "false"
                        ),
                        "call_to_actions": menu,
                    }
                ],
            }

            res = requests.post(
                self.url + "/custom_user_settings",
                json=dataJSON,
                headers=header,
                params=params,
            )
            return self.__analyse(res)

    @retry(requests.exceptions.ConnectionError, tries=3, delay=3)
    def send_file(
        self,
        sender_id,
        file,
        filetype="file",
        filename=None,
        reusable=False,
        **kwargs,
    ):
        """
        this method send an attachment from file

        Args:
            destId (str): user id facebook for the destination
            file (str): name of the file in local folder
            filetype (str, optional): type of the file["video","image",...]. Defaults to "file".
            filename (str, optional): A filename received for de destination . Defaults to None.
            reusable (bool, default False): Make an attachment reusable

        Returns:
            Response: POST request to the facebook API to send a file to the user

        Ref:
            https://developers.facebook.com/docs/messenger-platform/send-messages#file
        """
        if filename is None:
            filename = file
        params = {"access_token": self.token}

        dataJSON = {
            "recipient": json.dumps({"id": sender_id}),
            "message": json.dumps(
                {
                    "attachment": {
                        "type": filetype,
                        "payload": {
                            "is_reusable": reusable,
                        },
                    }
                }
            ),
            "filedata": (
                os.path.basename(filename),
                open(file, "rb"),
                f"{filetype}/{file.split('.')[-1]}",
            ),
        }
        dataJSON.update(kwargs)

        # multipart encode the entire payload
        multipart_data = requests_toolbelt.MultipartEncoder(dataJSON)

        # multipart header from multipart_data
        header = {"Content-Type": multipart_data.content_type}

        res = requests.post(
            self.url + "/messages",
            params=params,
            headers=header,
            data=multipart_data,
        )
        return self.__analyse(res)

    @retry(requests.exceptions.ConnectionError, tries=3, delay=3)
    def send_media(self, sender_id, fb_url, media_type, **kwargs):
        """
        Method that sends files media as image and video via facebook link.
        This model does not allow any external URLs, only those on Facebook.

        Args:
            destId (str): user id facebook for the destination
            fb_url (str): url of the media to send on facebook
                for this: To get the Facebook URL of an image or video, follow these steps:
                    -Click on the image or video thumbnail to open the full-size view
                    -Copy the URL address from your browser's address bar.
            media_type (str): the type of the media who to want send, available["image","video"]

        Returns:
            Response: POST request to the facebook API to send a media file using url facebook

        Ref:
            https://developers.facebook.com/docs/messenger-platform/send-messages/template/media
        """
        recipient = (
            {"one_time_notif_token": kwargs.get("one_time_notif_token")}
            if kwargs.get("one_time_notif_token")
            else {"id": sender_id}
        )

        dataJSON = {
            "recipient": recipient,
            "message": {
                "attachment": {
                    "type": "template",
                    "payload": {
                        "template_type": "media",
                        "elements": [{"media_type": media_type, "url": fb_url}],
                    },
                }
            },
        }
        dataJSON.update(kwargs)

        header = {"content-type": "application/json; charset=utf-8"}
        params = {"access_token": self.token}

        res = requests.post(
            f"https://graph.facebook.com/{self.api_version}/me/messages",
            json=dataJSON,
            headers=header,
            params=params,
        )
        return self.__analyse(res)

    @retry(requests.exceptions.ConnectionError, tries=3, delay=3)
    def get_started(self, payload="/"):
        """
        Method that GET STARTED button
        when the user talk first to the bot.

        Returns:
            Response: POST request to the facebook API to send a media file using url facebook

        Ref:
            https://developers.facebook.com/docs/messenger-platform/reference/messenger-profile-api/get-started-button
        """

        dataJSON = {"get_started": {"payload": payload}}

        header = {"content-type": "application/json; charset=utf-8"}
        params = {"access_token": self.token}

        res = requests.post(
            f"https://graph.facebook.com/{self.api_version}/me/messenger_profile",
            json=dataJSON,
            headers=header,
            params=params,
        )
        return self.__analyse(res)

    @retry(requests.exceptions.ConnectionError, tries=3, delay=3)
    def send_button(self, sender_id, buttons, text, **kwargs):
        """
        The button template sends a text message with
        up to three buttons attached. This template gives
        the message recipient different options to choose from,
        such as predefined answers to questions or actions to take.

         Args:
             sender_id (str): user id facebook for the destination
             buttons (list of dict) || (list of Button): The list of buttons who want send
             text (str): A text to describe the fonctionnality of the buttons

         Returns:
             Response: POST request to the facebook API to send all different buttons

         Ref:
             https://developers.facebook.com/docs/messenger-platform/send-messages/template/button
        """
        recipient = (
            {"one_time_notif_token": kwargs.get("one_time_notif_token")}
            if kwargs.get("one_time_notif_token")
            else {"id": sender_id}
        )

        buttons = [
            button.value if isinstance(button, Button) else button for button in buttons
        ]

        self.send_action(sender_id, "typing_on")
        dataJSON = {
            "recipient": recipient,
            "message": {
                "attachment": {
                    "type": "template",
                    "payload": {
                        "template_type": "button",
                        "text": text,
                        "buttons": buttons,
                    },
                }
            },
        }
        dataJSON.update(kwargs)

        header = {"content-type": "application/json; charset=utf-8"}
        params = {"access_token": self.token}

        self.send_action(sender_id, "typing_off")
        res = requests.post(
            self.url + "/messages",
            json=dataJSON,
            headers=header,
            params=params,
        )
        return self.__analyse(res)

    @retry(requests.exceptions.ConnectionError, tries=3, delay=3)
    def create_personas(self, name, profile_picture_url, **kwargs):
        """
        Method that creates a persona for the bot.

        Args:
            name (str): The name of the persona
            profile_picture_url (str): The url of the profile picture

        Returns:
            Response: POST request to the facebook API to create a persona

        Ref:
            https://developers.facebook.com/docs/messenger-platform/send-messages/personas
        """
        dataJSON = {
            "name": name,
            "profile_picture_url": profile_picture_url,
        }
        dataJSON.update(kwargs)
        header = {"content-type": "application/json; charset=utf-8"}
        params = {"access_token": self.token}

        res = requests.post(
            f"https://graph.facebook.com/{self.page_id}/personas",
            json=dataJSON,
            headers=header,
            params=params,
        )
        res = self.__analyse(res)
        return res.json().get("id")

    @retry(requests.exceptions.ConnectionError, tries=3, delay=3)
    def list_personas(self):
        """
        Method that lists all personas for the bot.

        Returns:
            Response: GET request to the facebook API to list all personas

        Ref:
            https://developers.facebook.com/docs/messenger-platform/send-messages/personas
        """
        header = {"content-type": "application/json; charset=utf-8"}
        params = {"access_token": self.token}

        res = requests.get(
            f"https://graph.facebook.com/{self.api_version}/{self.page_id}/personas",
            headers=header,
            params=params,
        )
        res = self.__analyse(res)
        return res.json().get("data")

    @retry(requests.exceptions.ConnectionError, tries=3, delay=3)
    def get_personas(self, persona_id):
        """
        Method that gets a persona for the bot.

        Args:
            persona_id (str): The id of the persona

        Returns:
            Response: GET request to the facebook API to get a persona

        Ref:
            https://developers.facebook.com/docs/messenger-platform/send-messages/personas
        """
        header = {"content-type": "application/json; charset=utf-8"}
        params = {"access_token": self.token}

        res = requests.get(
            f"https://graph.facebook.com/{self.api_version}/{persona_id}",
            headers=header,
            params=params,
        )
        res = self.__analyse(res)
        return res.json()

    @retry(requests.exceptions.ConnectionError, tries=3, delay=3)
    def delete_personas(self, persona_id):
        """
        Method that deletes a persona for the bot.

        Args:
            persona_id (str): The id of the persona

        Returns:
            Response: DELETE request to the facebook API to delete a persona

        Ref:
            https://developers.facebook.com/docs/messenger-platform/send-messages/personas
        """
        header = {"content-type": "application/json; charset=utf-8"}
        params = {"access_token": self.token}

        res = requests.delete(
            f"https://graph.facebook.com/{self.api_version}/{persona_id}",
            headers=header,
            params=params,
        )
        res = self.__analyse(res)
        return res

    @retry(requests.exceptions.ConnectionError, tries=3, delay=3)
    def send_receipt_template(
        self,
        sender_id,
        recipient_name,
        order_number,
        payment_method,
        receipt_elements,
        summary,
        currency="MGA",
        address=None,
        adjustments=None,
        order_url=None,
        timestamp=None,
        **kwargs,
    ):
        """
        Method that sends a receipt template to a customer  .

        Args:
            recipient_name (str | required ): The name of the recipient
            order_number (str | required ): The order number
            payment_method (str | required ): The payment method
            summary (Summary obj or dict | required ): The summary of the order
            currency (str): The currency of the order
            address (Adresse obj or dict | optional ): The address of the recipient
            adjustments (list of Adjustment obj or dict | optional ): The adjustments of the order
            order_url (str | optionnal ): The url of the order
            timestamp (str | optionnal ): The timestamp of the order

        Returns:
            Response: POST request to the facebook API to send a receipt template

        Ref:
            https://developers.facebook.com/docs/messenger-platform/reference/templates/receipt
        """
        recipient = (
            {"one_time_notif_token": kwargs.get("one_time_notif_token")}
            if kwargs.get("one_time_notif_token")
            else {"id": sender_id}
        )

        if receipt_elements:
            receipt_elements = [
                receipt.value if isinstance(receipt, ReceiptElement) else receipt
                for receipt in receipt_elements
            ]

        summary = summary.value if isinstance(summary, Summary) else summary
        address = address.value if isinstance(address, Address) else address

        if adjustments:
            adjustments = [
                adjustment.value if isinstance(adjustment, Adjustment) else adjustment
                for adjustment in adjustments
            ]

        dataJSON = {
            "recipient": recipient,
            "message": {
                "attachment": {
                    "type": "template",
                    "payload": {
                        "template_type": "receipt",
                        "recipient_name": recipient_name,
                        "order_number": order_number,
                        "currency": currency,
                        "payment_method": payment_method,
                        "address": address,
                        "summary": summary,
                        "elements": receipt_elements,
                        "order_url": order_url,
                        "adjustments": adjustments,
                        "timestamp": timestamp,
                    },
                }
            },
        }
        header = {"content-type": "application/json; charset=utf-8"}
        params = {"access_token": self.token}

        res = requests.post(
            self.url + "/messages",
            json=dataJSON,
            headers=header,
            params=params,
        )
        return self.__analyse(res)

    @retry(requests.exceptions.ConnectionError, tries=3, delay=3)
    def get_user_profile(self, sender_id, fields="first_name,last_name,profile_pic"):
        """
        The User Profile methiod allows you to use a Page-scoped ID (PSID)
        to retrieve user profile information that can be used to personalize
        the experience of people interacting with your Messenger.
        """
        params = {"fields": fields, "access_token": self.token}
        res = requests.get(f"https://graph.facebook.com/{sender_id}", params=params)
        res = self.__analyse(res)
        return res.json() if res.status_code == 200 else {}

    @retry(requests.exceptions.ConnectionError, tries=3, delay=3)
    def send_recurring_notification_optin(self, sender_id, optin, **kwargs):
        """
        Method that sends a recurring notification optin to a customer  .

        Args:
            optin (Optin obj or dict | required ): The optin

        Returns:
            Response: POST request to the facebook API to send a recurring notification optin

        Ref:
            https://developers.facebook.com/docs/messenger-platform/send-messages/recurring-notifications
        """

        optin = optin.value if isinstance(optin, RecurringNotificationOptin) else optin
        dataJSON = {
            "recipient": {"id": sender_id},
            "message": {"attachment": {"type": "template", "payload": optin}},
        }

        header = {"content-type": "application/json; charset=utf-8"}
        params = {"access_token": self.token}

        res = requests.post(
            self.url + "/messages",
            json=dataJSON,
            headers=header,
            params=params,
        )
        return self.__analyse(res)

    @retry(requests.exceptions.ConnectionError, tries=3, delay=3)
    def send_recurring_notification(self, message_token, elements, **kwargs):
        """
        Method that sends a recurring notification to a customer .

        Args:
            message_token (str | required ): The message token from optin
            elements (list of Element obj or list of dict | required ): The recurring notification elements

        Returns:
            Response: POST request to the facebook API to send a recurring notification

        Ref:
            https://developers.facebook.com/docs/messenger-platform/send-messages/recurring-notifications
        """

        elements = [
            element.value if isinstance(element, Element) else element
            for element in elements
        ]

        dataJSON = {
            "recipient": {"notification_messages_token": message_token},
            "message": {
                "attachment": {
                    "type": "template",
                    "payload": {
                        "template_type": "notification_messages",
                        "elements": elements,
                    },
                }
            },
        }

        header = {"content-type": "application/json; charset=utf-8"}
        params = {"access_token": self.token}

        res = requests.post(
            self.url + "/messages",
            json=dataJSON,
            headers=header,
            params=params,
        )

        return self.__analyse(res)

    @retry(requests.exceptions.ConnectionError, tries=3, delay=3)
    def send_onetime_notification_request(self, sender_id, title, payload, **kwargs):
        """
        Method to send a notification request.

        Args:
            sender_id (str | required ): The customer id
            title (str | optional ): The title of the notification request

        Returns:
            Response: POST request to the facebook API to send a notification request

        Ref:
            https://developers.facebook.com/docs/messenger-platform/send-messages/one-time-notification
        """

        payload = (
            Payload.trt_payload_out(payload)
            if isinstance(payload, Payload)
            else payload
        )

        dataJson = {
            "recipient": {"id": sender_id},
            "message": {
                "attachment": {
                    "type": "template",
                    "payload": {
                        "template_type": "one_time_notif_req",
                        "title": title or "Notification request",
                        "payload": payload,
                    },
                }
            },
        }

        header = {"content-type": "application/json; charset=utf-8"}
        params = {"access_token": self.token}

        res = requests.post(
            self.url + "/messages",
            json=dataJson,
            headers=header,
            params=params,
        )
        return self.__analyse(res)

    @retry(requests.exceptions.ConnectionError, tries=3, delay=3)
    def send_product_template(
        self, sender_id, products, next=None, quick_rep=None, **kwargs
    ):
        """
        Method that sends a product template to a customer owned by the page.
        Args:
            sender_id (str | required ): The id of the customer
            products (list of id  | required ): The products ids to send
            next(bool) || (text): this params activate the next page when elements have a length more than ten
        Returns:
            Response: POST request to the facebook API to send a product template
        Ref:
            https://developers.facebook.com/docs/messenger-platform/send-messages/template/product
        """

        products = [
            product.value if isinstance(product, Product) else product
            for product in products
        ]

        dataJSON = {
            "recipient": {"id": sender_id},
            "message": {
                "attachment": {
                    "type": "template",
                    "payload": {
                        "template_type": "product",
                        "elements": products[:10],
                    },
                }
            },
        }
        dataJSON.update(kwargs)

        if len(products) > 10 and next:
            dataJSON["message"]["quick_replies"] = [
                {
                    "content_type": "text",
                    "title": "Next" if next == True else str(next),
                    "payload": "/__next",
                    "image_url": "https://icon-icons.com/downloadimage.php"
                    + "?id=81300&root=1149/PNG/512/&file="
                    + "1486504364-chapter-controls-forward-play"
                    + "-music-player-video-player-next_81300.png",
                }
            ]
            pickle.dump(
                (products[10:], next),
                open(f"assets/private/.__{sender_id}", "wb"),
            )

        if quick_rep:
            quick_rep = [
                qr.value if isinstance(qr, QuickReply) else qr for qr in quick_rep
            ]

            if isinstance(dataJSON["message"].get("quick_replies"), list):
                dataJSON["message"]["quick_replies"] = (
                    quick_rep + dataJSON["message"]["quick_replies"]
                )
            else:
                dataJSON["message"]["quick_replies"] = quick_rep

        header = {"content-type": "application/json; charset=utf-8"}
        params = {"access_token": self.token}

        res = requests.post(
            self.url + "/messages",
            json=dataJSON,
            headers=header,
            params=params,
        )
        return self.__analyse(res)



from .tools import funcs


def command(*args, **kwargs):
    """
    A decorator that registers the function as the route
        of a processing per command sent.
    """

    def call_fn(function):
        funcs["command"][args[0]] = function

    return call_fn


def action(*args, **kwargs):
    """
    A decorator that registers the function as the route
        of a defined action handler.
    """

    def call_fn(function):
        funcs["action"][args[0]] = function

    return call_fn


def event(*args, **kwargs):
    """
    A decorator that registers the function as the route
        of a defined event handler.
    """

    def call_fn(function):
        funcs["event"][args[0]] = function

    return call_fn


def before_receive(*args, **kwargs):
    """
    A decorator that run the function before
        running apropriate function
    """

    def call_fn(function):
        funcs["before"] = function

    return call_fn


def after_receive(*args, **kwargs):
    """
    A decorator that run the function after
        running apropriate function
    """

    def call_fn(function):
        funcs["after"] = function

    return call_fn



"""
    List of All UI Widget Messenger 
"""
from .payload import Payload
from .constant import (
    Type,
    Content_type,
    Notification_reoptin,
    Notification_cta_text,
    Notification_frequency,
)


class QuickReply:
    def __init__(self, **kwargs):
        """
        Object that can be used to generated a quick_reply
        """
        self.content_type = kwargs.get("content_type", "text")
        if self.content_type not in ("text", "user_phone_number", "user_email"):
            raise ValueError(
                "content_type can only be 'text', 'user_phone_number', 'user_email'"
            )

        self.title = kwargs.get("title")
        self.payload = kwargs.get("payload")

        if self.content_type == "text" and not self.payload:
            raise ValueError("payload must be present for text")

        if self.content_type == "text" and not self.title:
            raise ValueError("title must be present for text")

        self.image_url = kwargs.get("image_url")

    @property
    def value(self):
        res = {"content_type": self.content_type}

        if self.content_type == "text":
            res["title"] = self.title
            res["payload"] = Payload.trt_payload_out(self.payload)

            if self.image_url:
                res["image_url"] = self.image_url
        return res

    def __str__(self):
        return str(self.value)


class Button:
    def __init__(self, **kwargs):

        self.type = kwargs.get("type", "postback")
        self.title = kwargs.get("title")
        self.url = kwargs.get("url")
        self.payload = kwargs.get("payload")

        if self.type not in (
            "postback",
            "web_url",
            "phone_number",
            "account_link",
            "account_unlink",
        ):
            raise ValueError(
                "type must be one of 'postback', 'web_url', 'phone_number', 'account_link', 'account_unlink'"
            )

        if self.type in ("postback", "phone_number"):

            if not self.payload:
                raise ValueError("payload must be present")

            if not self.title:
                raise ValueError("title must be present")

        elif self.type == "web_url":
            if not self.url:
                raise ValueError("url must be present for web_url")

            if not self.title:
                raise ValueError("title must be present for web_url")

        elif self.type == "account_link":
            if not self.url:
                raise ValueError("url facebook login must be present for account_link")

    @property
    def value(self):
        if self.payload:
            self.payload = Payload.trt_payload_out(self.payload)

        if self.type in ("postback", "phone_number"):
            return {"type": self.type, "title": self.title, "payload": self.payload}

        elif self.type == "web_url":
            return {"type": "web_url", "title": self.title, "url": self.url}

        elif self.type == "account_link":
            return {"type": "account_link", "url": self.url}

        elif self.type == "account_unlink":
            return {"type": "account_unlink"}

    def __str__(self):
        return str(self.value)


class Element:
    def __init__(self, **kwargs):
        self.title = kwargs.get("title")
        self.subtitle = kwargs.get("subtitle")
        self.image = kwargs.get("image_url")
        self.buttons = kwargs.get("buttons")
        self.default_action = kwargs.get("default_action")

        if not self.title:
            raise ValueError("Element must be have a title")

        if not self.buttons:
            raise ValueError("Element must be have a buttons")

        if not isinstance(self.buttons, list):
            raise ValueError("buttons must be a list of Button")

        if len(self.buttons) > 3:
            raise ValueError("buttons must be three maximum")

        for i in range(len(self.buttons)):
            if not isinstance(self.buttons[i], Button):
                raise ValueError("buttons must a List of Button")

        if self.default_action:
            if not isinstance(self.default_action, Button):
                raise ValueError("default_action must be a Button")

    @property
    def value(self):
        res = {"title": self.title}

        if self.subtitle:
            res["subtitle"] = self.subtitle

        if self.image:
            res["image_url"] = self.image

        if self.default_action:
            res["default_action"] = self.default_action.value

        res["buttons"] = [button.value for button in self.buttons]  # type: ignore

        return res

    def __str__(self):
        return str(self.value)


class ReceiptElement:
    def __init__(self, **kwargs):
        self.title = kwargs.get("title")
        self.subtitle = kwargs.get("subtitle")
        self.quantity = kwargs.get("quantity")
        self.price = kwargs.get("price")
        self.currency = kwargs.get("currency")
        self.image = kwargs.get("image_url")

        if not self.title:
            raise ValueError("Receipt element must be have a title")

        if not self.price:
            raise ValueError("Receipt element must be have a price")

    @property
    def value(self):
        res = {"title": self.title}

        if self.subtitle:
            res["subtitle"] = self.subtitle

        if self.quantity:
            res["quantity"] = self.quantity

        res["price"] = self.price
        res["currency"] = self.currency

        if self.image:
            res["image_url"] = self.image

        return res

    def __str__(self):
        return str(self.value)


class Summary:
    def __init__(self, **kwargs):
        self.subtotal = kwargs.get("subtotal")
        self.shipping_cost = kwargs.get("shipping_cost")
        self.total_tax = kwargs.get("total_tax")
        self.total_cost = kwargs.get("total_cost")

        if not self.total_cost:
            raise ValueError("Summary must have a total_cost")

    @property
    def value(self):
        res = {"total_cost": self.total_cost}

        if self.shipping_cost:
            res["shipping_cost"] = self.shipping_cost

        if self.total_tax:
            res["total_tax"] = self.total_tax

        if self.subtotal:
            res["subtotal"] = self.subtotal

        return res

    def __str__(self):
        return str(self.value)


class Address:
    def __init__(self, **kwargs):
        self.street_1 = kwargs.get("street_1")
        self.street_2 = kwargs.get("street_2")
        self.city = kwargs.get("city")
        self.postal_code = kwargs.get("postal_code")
        self.state = kwargs.get("state")
        self.country = kwargs.get("country")

        if not self.street_1:
            raise ValueError("Address must have a street_1")

        if not self.city:
            raise ValueError("Address must have a city")

        if not self.postal_code:
            raise ValueError("Address must have a postal_code")

        if not self.state:
            raise ValueError("Address must have a state")

        if not self.country:
            raise ValueError("Address must have a country")

    @property
    def value(self):
        res = {}

        if self.street_1:
            res["street_1"] = self.street_1

        if self.street_2:
            res["street_2"] = self.street_2

        if self.city:
            res["city"] = self.city

        if self.postal_code:
            res["postal_code"] = self.postal_code

        if self.state:
            res["state"] = self.state

        if self.country:
            res["country"] = self.country

        return res

    def __str__(self):
        return str(self.value)


class Adjustment:
    def __init__(self, **kwargs):
        self.name = kwargs.get("name")
        self.amount = kwargs.get("amount")

        if not self.name:
            raise ValueError("Adjustment must be have a name")

        if not self.amount:
            raise ValueError("Adjustment must be have a amount")

    @property
    def value(self):
        return {"name": self.name, "amount": self.amount}

    def __str__(self):
        return str(self.value)


class RecurringNotificationOptin:
    def __init__(self, **kwargs):
        """
        RecurringNotificationOptin ui object to be used in the send_optin
        Args:
            title (str) - Title of the optin
            image (str | optional) - Image url of the optin
            payload (str) - Payload of the optin
            notification_frequency (str) - Frequency of the notification messages
        """
        self.template_type = "notification_messages"
        self.title = kwargs.get("title")
        self.image_url = kwargs.get("image_url")
        self.payload = kwargs.get("payload")
        self.notification_frequency = kwargs.get("notification_frequency")
        self.notification_cta_text = kwargs.get("notification_cta_text")
        self.notification_reoptin = kwargs.get("notification_reoptin")

        if not self.title:
            raise ValueError("RecurringNotificationOptin must have a title")

        if not self.payload:
            raise ValueError("RecurringNotificationOptin must have a payload")

        if not self.notification_frequency:
            raise ValueError(
                "RecurringNotificationOptin must have a notification frequency"
            )

        if self.notification_frequency not in Notification_frequency:
            raise ValueError(
                "RecurringNotificationOptin must have a valid notification frequency"
            )

        if self.notification_cta_text:
            if self.notification_cta_text not in Notification_cta_text:
                raise ValueError(
                    "RecurringNotificationOptin must have a valid notification cta text"
                )

        if self.notification_reoptin:
            if self.notification_reoptin not in Notification_reoptin:
                raise ValueError(
                    "RecurringNotificationOptin must have a valid notification reoptin"
                )

    @property
    def value(self):
        res = {
            "template_type": self.template_type,
            "title": self.title,
            "notification_messages_frequency": self.notification_frequency.value,  # type: ignore
        }

        if isinstance(self.payload, Payload):
            res["payload"] = Payload.trt_payload_out(self.payload)
        else:
            res["payload"] = self.payload

        if self.image_url:
            res["image_url"] = self.image_url

        if self.notification_cta_text:
            res["notification_messages_cta_text"] = self.notification_cta_text.value

        if self.notification_reoptin:
            res["notification_messages_reoptin"] = self.notification_reoptin.value

        return res

    def __str__(self):
        return str(self.value)


class Product:
    def __init__(self, id):
        self.id = id

    @property
    def value(self):
        return {"id": self.id}

    def __str__(self):
        return str(self.value)



import os
import sys
import time
import inspect
import colorama
import tempfile
from . import source

__version__ = "1.1.8"
__author__ = "iTeam-$"


colorama.init()


def typing_print(text):
    for character in text:
        sys.stdout.write(character)
        sys.stdout.flush()
        time.sleep(0.02)
    print()


def create_env(path):
    if sys.platform == "win32":
        print(source.ENV_CMD, file=open(f"{path}/.env.bat", "w"))
    else:
        print(source.ENV, file=open(f"{path}/.env", "w"))
    typing_print("~\033[32m 👌 \033[0m | Env file created")


def create_lang(path):
    print(source.LANGS, file=open(f"{path}/langs.json", "w"))
    typing_print("~\033[32m 👌 \033[0m | Langs file created")


def init_proj(path):
    create_env(path)
    create_lang(path)
    print(source.CORE, file=open(f"{path}/core.py", "w"))
    typing_print("~\033[32m 👌 \033[0m | Core file created")

    print(source.CONF, file=open(f"{path}/conf.py", "w"))
    typing_print("~\033[32m 👌 \033[0m | Config file created")

    for folder in {"public", "private"}:
        os.makedirs(os.path.join(path, "assets", folder), exist_ok=True)

    print(
        ".env\n.env.bat\n__pycache__/\nngrok\nngrok.exe\n_db.json",
        file=open(f"{path}/.gitignore", "a"),
    )

    print(
        "ampalibe",
        file=open(f"{path}/requirements.txt", "a"),
    )


if sys.argv[0] == "-m" and len(sys.argv) > 1:
    if sys.argv[1] == "version":
        typing_print("\033[32m" + __version__ + " ⭐ \033[0m")

    elif sys.argv[1] == "init":
        typing_print("~\033[32m 👌 \033[0m | Initiating  ...")
        init_proj(".")
        typing_print(
            inspect.cleandoc(
                """
                ~\033[32m 👌 \033[0m | Project Ampalibe initiated. \033[32mYoupii !!! 😎 \033[0m
                ~\033[36m TIPS\033[0m |\033[0m Fill in .env file.
                ~\033[36m TIPS\033[0m |\033[36m ampalibe run\033[0m for lauching project.
            """
            )
        )

    elif sys.argv[1] == "create":
        proj_name = sys.argv[2]
        typing_print(f"~\033[32m 👌 \033[0m | Creating {proj_name} ...")
        os.makedirs(proj_name)
        init_proj(proj_name)
        typing_print(
            inspect.cleandoc(
                f"""
                ~\033[32m 👌 \033[0m | Project Ampalibe created. \033[32mYoupii !!! 😎 \033[0m
                ~\033[36m TIPS\033[0m |\033[0m Fill in .env file.
                ~\033[36m TIPS\033[0m |\033[36m cd {proj_name} && ampalibe run\033[0m for lauching project.
            """
            )
        )

    elif sys.argv[1] == "env":
        create_env(".")

    elif sys.argv[1] == "lang":
        create_lang(".")

    elif sys.argv[1] == "run":
        print(
            inspect.cleandoc(
                "\033[36m"
                + r"""
                                                                0o
                                                                Oo
                                                                coooool
                                                               looooooool
                                                              loooooooooool
            _    __  __ ____   _    _     ___ ____  _____     looooooooooool
           / \  |  \/  |  _ \ / \  | |   |_ _| __ )| ____|    looooooooooool
          / _ \ | |\/| | |_) / _ \ | |    | ||  _ \|  _|       loooooooooool
         / ___ \| |  | |  __/ ___ \| |___ | || |_) | |___        looooooool
        /_/   \_\_|  |_|_| /_/   \_\_____|___|____/|_____|         oooooo
        """
                + "\033[0m"
            )
        )
        typing_print(
            "~\033[32m 👌\033[0m | Env Loaded\n~\033[32m 👌\033[0m | Ampalibe"
            " running..."
        )

    elif sys.argv[1] == "usage":
        typing_print(
            inspect.cleandoc(
                """
                    Usage: ampalibe \033[32m { create, init, env, run, version, help } \033[0m
                    ------
                    👉 \033[32m create ... : \033[0m create a new project in a new directory specified
                    👉 \033[32m init: \033[0m create a new project in current dir
                    👉 \033[32m version: \033[0m show the current version
                    👉 \033[32m env: \033[0m generate only a .env file
                    👉 \033[32m lang: \033[0m generate only a langs.json file
                    👉 \033[32m run [--dev]: \033[0m run the server, autoreload if --dev is specified
                    👉 \033[32m help: \033[0m show this current help
                """
            )
        )

    sys.exit(0)

try:
    from conf import Configuration  # type: ignore
except ImportError:
    dir_tmp = os.path.join(tempfile.gettempdir(), "ampalibe_temp")
    os.makedirs(dir_tmp, exist_ok=True)
    with open(os.path.join(dir_tmp, "conf.py"), "w") as f:
        f.write(source.CONF)
    sys.path.insert(0, dir_tmp)
finally:
    from .constant import *
    from .model import Model
    from .logger import Logger
    from aiocron import crontab
    from .payload import Payload
    from .messenger import Messenger
    from .core import webserver, Init as init
    from .utils import translate, download_file, simulate
    from .decorators import (
        event,
        action,
        command,
        before_receive,
        after_receive,
    )



ENV = """# PAGE ACCESS TOKEN 
export AMP_ACCESS_TOKEN=

# PAGE VERIF TOKEN
export AMP_VERIF_TOKEN= 


# DATABASE AUTHENTIFICATION
export ADAPTER=SQLITE
#export ADAPTER=MYSQL
#export ADAPTER=POSTGRESQL
#export ADAPTER=MONGODB

####### CASE MYSQL, POSTGRESQL OR MONGODB ADAPTER
export DB_HOST=
export DB_USER=
export DB_PASSWORD=
export DB_NAME=
#export DB_PORT=
#export SRV_PROTOCOL=1

####### CASE SQLITE ADAPTER
export DB_FILE=ampalibe.db


# APPLICATION CONFIGURATION
export AMP_HOST=0.0.0.0
export AMP_PORT=4555

# URL APPLICATION
export AMP_URL=

"""

ENV_CMD = """:: PAGE ACCESS TOKEN 
set AMP_ACCESS_TOKEN=

:: PAGE VERIF TOKEN
set AMP_VERIF_TOKEN= 

:: DATABASE AUTHENTIFICATION
set ADAPTER=SQLITE
:: set ADAPTER=MYSQL
:: set ADAPTER=POSTGRESQL
:: set ADAPTER=MONGODB

::::: CASE MYSQL, POSTGRESQL OR MONGODB ADAPTER
set DB_HOST=
set DB_USER=
set DB_PASSWORD=
set DB_NAME=
:: set DB_PORT=
:: set SRV_PROTOCOL=1

:: CASE SQLITE ADAPTER
set DB_FILE=ampalibe.db


:: APPLICATION CONFIGURATION
set AMP_HOST=0.0.0.0
set AMP_PORT=4555

:: URL APPLICATION
set AMP_URL=

"""

CORE = """import ampalibe
from ampalibe import Messenger

chat = Messenger()

# create a get started option to get permission of user.
# chat.get_started()

@ampalibe.command('/')
def main(sender_id, cmd, **ext):
    '''
    main function where messages received on
    the facebook page come in.

    @param sender_id String: 
        sender facebook id
    @param cmd String: 
        message content
    @param ext Dict: 
        contain list of others
            data sent by facebook (sending time, ...)
            data sent by your payload if not set in parameter
    '''
    
    chat.send_text(sender_id, "Hello, Ampalibe")
    """

CONF = """from os import environ as env


class Configuration:
    '''
        Retrieves the value from the environment.
        Takes the default value if not defined.
    '''
    ADAPTER = env.get('ADAPTER')

    DB_FILE = env.get('DB_FILE')
    
    DB_HOST = env.get('DB_HOST', 'localhost')
    DB_USER = env.get('DB_USER', 'root')
    DB_PASSWORD = env.get('DB_PASSWORD', '')
    DB_PORT = env.get('DB_PORT')
    DB_NAME = env.get('DB_NAME')
    SRV_PROTOCOL = env.get('SRV_PROTOCOL')

    ACCESS_TOKEN = env.get('AMP_ACCESS_TOKEN')
    VERIF_TOKEN = env.get('AMP_VERIF_TOKEN')

    APP_HOST = env.get('AMP_HOST', '0.0.0.0')
    APP_PORT = int(env.get('AMP_PORT', 4555))
    APP_URL = env.get('AMP_URL')
    
"""

LANGS = """{
    "hello_world": {
        "en": "Hello World",
        "fr": "Bonjour le monde"
    },

    "ampalibe": {
        "en": "Jackfruit", 
        "fr": "Jacquier",
        "mg": "Ampalibe"
    }
}
"""



import setuptools

with open("README.md", "r") as fh:
    long_description = fh.read()

setuptools.setup(
    name="ampalibe",  # This is the name of the package
    version="1.1.8",  # The release version
    author="iTeam-$",  # Full name of the author
    description=(
        "Ampalibe is a lightweight Python framework for building Facebook"
        " Messenger bots faster."
    ),
    long_description=long_description,  # Long description read from the readme
    long_description_content_type="text/markdown",
    packages=["ampalibe"],  # List of all modules to be installed
    classifiers=[
        "Programming Language :: Python :: 3",
        "Programming Language :: Python :: 3.7",
        "Programming Language :: Python :: 3.8",
        "Programming Language :: Python :: 3.9",
        "Programming Language :: Python :: 3.10",
        "Programming Language :: Python :: 3.11",
        "License :: OSI Approved :: MIT License",
        "Operating System :: OS Independent",
    ],  # Information to filter the project on PyPi website
    python_requires=">=3.7",  # Name of the python package
    install_requires=[
        "fastapi",
        "uvicorn",
        "retry",
        "requests",
        "colorama",
        "requests_toolbelt",
        "watchdog!=2.2.0",
        "aiocron",
        "tinydb",
    ],  # depandance
    include_package_data=True,  # Include all data file with the package
    scripts=["bin/ampalibe", "bin/ampalibe.bat"],
)



import sys
import logging


class CustomFormatter(logging.Formatter):
    """
    Custom formatter for logging, add colors to the output and custom format
    """

    grey = "\x1b[38;20m"
    yellow = "\x1b[33;20m"
    red = "\x1b[31;20m"
    green = "\x1b[32;20m"
    bold_red = "\x1b[31;1m"
    reset = "\x1b[0m"
    _format = "%(levelname)s|%(name)s:   %(message)s "

    # Define format of each status  (DEBUG|INFO|WARNINGS|ERROR|CRITICAL)
    FORMATS = {
        logging.DEBUG: grey + _format + reset,
        logging.INFO: green + _format + reset,
        logging.WARNING: yellow + _format + reset,
        logging.ERROR: red + _format + reset,
        logging.CRITICAL: bold_red + _format + reset,
    }

    def format(self, record):
        log_fmt = self.FORMATS.get(record.levelno)
        formatter = logging.Formatter(log_fmt)
        return formatter.format(record)


def __logger(name="Ampalibe"):

    log = logging.getLogger(name)
    log.setLevel(logging.DEBUG)
    # by default the logger will print on the console , in stdout
    handler = logging.StreamHandler(sys.stdout)
    handler.setLevel(logging.DEBUG)
    handler.setFormatter(CustomFormatter())
    log.addHandler(handler)
    return log


Logger = __logger()



# pyright: reportGeneralTypeIssues=false

import os
from .payload import Payload
from datetime import datetime
from conf import Configuration  # type: ignore
from tinydb import TinyDB, Query
from tinydb.operations import delete


class DataBaseConfig:
    def standart(self, conf=Configuration):
        """
        function to configure the standart database
        """
        db_conf = {
            "host": conf.DB_HOST,
            "user": conf.DB_USER,
            "password": conf.DB_PASSWORD,
            "database": conf.DB_NAME,
        }
        if conf.DB_PORT:
            db_conf["port"] = conf.DB_PORT
        return db_conf

    def mongodb(self, conf=Configuration):
        """
        function to configure the mongodb database
        """

        url = "mongodb"
        if hasattr(conf, "SRV_PROTOCOL"):
            url += "+srv://" if not conf.SRV_PROTOCOL else "://"

        if conf.DB_USER and conf.DB_PASSWORD:
            url += f"{conf.DB_USER}:{conf.DB_PASSWORD}@"

        url += conf.DB_HOST

        if conf.DB_PORT:
            url += f":{str(conf.DB_PORT)}/"

        return url


class Model:
    """
    Object for interact with database with pre-defined function
    """

    def __init__(self, conf=Configuration, init=True):
        """
        object to interact with database

        @params: conf [ Configuration object ]
        @return: Request object
        """
        if not init:
            return
        self.ADAPTER = conf.ADAPTER

        if self.ADAPTER in ("MYSQL", "POSTGRESQL"):
            self.DB_CONF = DataBaseConfig().standart(conf)
        elif self.ADAPTER == "MONGODB":
            self.DB_CONF = DataBaseConfig().mongodb(conf)
        else:  # SQLite is choosen by default
            self.DB_CONF = conf.DB_FILE

        self.__connect()
        self.__init_db()
        os.makedirs("assets/private/", exist_ok=True)
        self.tinydb = TinyDB("assets/private/_db.json")

    def __connect(self):
        """
        The function which connect object to the database.
        """
        if self.ADAPTER == "MYSQL":
            try:
                import mysql.connector
            except ImportError:
                raise ImportError(
                    "You must install mysql-connector to use mysql database: pip install mysql-connector"
                )
            self.db = mysql.connector.connect(**self.DB_CONF)

        elif self.ADAPTER == "POSTGRESQL":
            try:
                import psycopg2
            except ImportError:
                raise ImportError(
                    "You must install psycopg2 to use postgresql database: pip install psycopg2"
                )
            self.db = psycopg2.connect(**self.DB_CONF)

        elif self.ADAPTER == "MONGODB":
            try:
                import pymongo
            except ImportError:
                raise ImportError(
                    "You must install pymongo to use mongodb database: pip install pymongo"
                )
            self.db = pymongo.MongoClient(self.DB_CONF)
            self.db = self.db[Configuration.DB_NAME]
            return  # no cursor for mongodb

        else:
            import sqlite3

            self.db = sqlite3.connect(
                self.DB_CONF,
                detect_types=sqlite3.PARSE_DECLTYPES | sqlite3.PARSE_COLNAMES,
            )

        self.cursor = self.db.cursor()

    def __init_db(self):
        """
        Creation of table if not exist
        Check the necessary table if exists
        """

        if self.ADAPTER == "MYSQL":
            req = """
                CREATE TABLE IF NOT EXISTS `amp_user` (
                    `id` INT NOT NULL AUTO_INCREMENT,
                    `user_id` varchar(50) NOT NULL UNIQUE,
                    `action` TEXT DEFAULT NULL,
                    `last_use` datetime NOT NULL DEFAULT current_timestamp(),
                    `lang` varchar(5) DEFAULT NULL,
                    PRIMARY KEY (`id`),
                    INDEX (last_use)
                ) ENGINE=InnoDB DEFAULT CHARSET=utf8mb4;
            """
        elif self.ADAPTER == "POSTGRESQL":
            req = """
                CREATE TABLE IF NOT EXISTS  "amp_user" (
                    id SERIAL,
                    user_id VARCHAR NULL DEFAULT NULL,
                    action TEXT NULL DEFAULT NULL,
                    last_use TIMESTAMP NULL DEFAULT NOW(),
                    lang VARCHAR NULL DEFAULT NULL,
                    PRIMARY KEY (id),
                    UNIQUE (user_id),
                    INDEX (last_use)
                )
            """
        elif self.ADAPTER == "MONGODB":
            if "amp_user" not in self.db.list_collection_names():
                self.db.create_collection("amp_user")
                # self.db.amp_user.create_index("user_id", unique=True)
            return
        else:
            req = """
               CREATE TABLE IF NOT EXISTS amp_user (
                   id INTEGER PRIMARY KEY AUTOINCREMENT, 
                   user_id TEXT NOT NULL UNIQUE,
                   action TEXT,
                   last_use TIMESTAMP DEFAULT CURRENT_TIMESTAMP,
                   lang TEXT
                )
            """
        self.cursor.execute(req)
        self.db.commit()

    def verif_db(fonction):  # type: ignore
        """
        decorator that checks if the database
        is connected or not before doing an operation.
        """

        def trt_verif(*arg, **kwarg):
            arg[0].__connect()
            return fonction(*arg, **kwarg)

        return trt_verif

    @verif_db
    def _verif_user(self, user_id):
        """
        method to insert new user and/or update the date
        of last use if the user already exists.

        @params :  user_id

        """
        SQL_MAP = {
            "MYSQL": "INSERT INTO amp_user(user_id) VALUES (%s) ON DUPLICATE KEY UPDATE last_use = NOW();",
            "POSTGRESQL": "INSERT INTO amp_user(user_id) VALUES (%s) ON CONFLICT (user_id) DO UPDATE SET last_use = NOW();",
            "SQLITE": "INSERT INTO amp_user(user_id) VALUES (?) ON CONFLICT(user_id) DO UPDATE SET last_use = CURRENT_TIMESTAMP;",
        }

        if self.ADAPTER == "MONGODB":
            self.db.amp_user.update_one(
                {"user_id": user_id},
                {"$set": {"last_use": datetime.now()}},
                upsert=True,
            )
        else:
            sql = SQL_MAP.get(self.ADAPTER)
            self.cursor.execute(sql, (user_id,))
            self.db.commit()

    @verif_db
    def get_action(self, sender_id):
        """
        get current action of an user

         @params :  sender_id
         @return : current action [ type of String/None ]
        """
        return self.get(sender_id, "action")[0]

    @verif_db
    def set_action(self, sender_id, action):
        """
        define a current action if an user

        @params :  sender_id, action
        @return:  None
        """
        if isinstance(action, Payload):
            action = Payload.trt_payload_out(action)

        if self.ADAPTER in ("MYSQL", "POSTGRESQL"):
            req = "UPDATE amp_user set action = %s WHERE user_id = %s"
        elif self.ADAPTER == "MONGODB":
            self.db.amp_user.update_one(
                {"user_id": sender_id},
                {"$set": {"action": action}},
            )
            return
        else:
            req = "UPDATE amp_user set action = ? WHERE user_id = ?"
        self.cursor.execute(req, (action, sender_id))
        self.db.commit()

    @verif_db
    def set_temp(self, sender_id, key, value):
        """
        set a temp parameter of an user

         @params:  sender_id
         @return:  None
        """
        if self.ADAPTER == "MONGODB":
            self.db.amp_user.update_one(
                {"user_id": sender_id},
                {"$set": {key: value}},
            )
            return
        if not self.tinydb.update({key: value}, Query().user_id == sender_id):
            self.tinydb.insert({"user_id": sender_id, key: value})

    @verif_db
    def get_temp(self, sender_id, key):
        """
        get one temporary data of an user

        @parmas :  sender_id
                   key
        @return: data
        """
        if self.ADAPTER == "MONGODB":
            return self.db.amp_user.find({"user_id": sender_id})[0].get(key)

        res = self.tinydb.search(Query().user_id == sender_id)
        if res:
            return res[0].get(key)

    @verif_db
    def del_temp(self, sender_id, key):
        """
        delete temporary parameter of an user

        @parameter :  sender_id
                      key
        @return: None
        """
        if self.ADAPTER == "MONGODB":
            self.db.amp_user.update_one(
                {"user_id": sender_id},
                {"$unset": {key: ""}},
            )
            return
        self.tinydb.update(delete(key), Query().user_id == sender_id)

    @verif_db
    def get_lang(self, sender_id):
        """
        get current lang of an user

        @params: sender_id
        @return lang or None
        """
        return self.get(sender_id, "lang")[0]

    @verif_db
    def set_lang(self, sender_id, lang):
        """
        define a current lang for an user

        @params :  sender_id
        @return:  None
        """
        if self.ADAPTER in ("MYSQL", "POSTGRESQL"):
            req = "UPDATE amp_user set lang = %s WHERE user_id = %s"
        elif self.ADAPTER == "MONGODB":
            self.db.amp_user.update_one(
                {"user_id": sender_id},
                {"$set": {"lang": lang}},
            )
            return
        else:
            req = "UPDATE amp_user set lang = ? WHERE user_id = ?"
        self.cursor.execute(req, (lang, sender_id))
        self.db.commit()

    @verif_db
    def get(self, sender_id, *args):
        """
        get specific data of an user

        @params :  sender_id, list of data to get
        @return:  list of data
        """
        if self.ADAPTER in ("MYSQL", "POSTGRESQL"):
            req = f"SELECT {','.join(args)} FROM amp_user WHERE user_id = %s"
        elif self.ADAPTER == "MONGODB":
            data = self.db.amp_user.find({"user_id": sender_id})[0]
            return [data.get(k) for k in args]
        else:
            req = f"SELECT {','.join(args)} FROM amp_user WHERE user_id = ?"
        self.cursor.execute(req, (sender_id,))
        return self.cursor.fetchone() or (None, None)



import os
import pickle
from .cmd import Cmd
from threading import Thread
from .messenger import Messenger

funcs = {
    "command": {},
    "action": {},
    "event": {},
    "before": None,
    "after": None,
}


def analyse(data):
    """
    Function analyzing data received from Facebook
    The data received are of type Json .
    """

    for event in data["entry"]:
        messaging = event["messaging"]

        for message in messaging:

            sender_id = message["sender"]["id"]

            if message.get("message"):

                if message["message"].get("attachments"):
                    # Get file name
                    data = message["message"].get("attachments")
                    # creation de l'objet cmd personalisé
                    atts = list(map(lambda dt: dt["payload"]["url"], data))
                    cmd = Cmd(atts[0])
                    cmd.set_atts(atts)
                    cmd.webhook = "attachments"
                    return sender_id, cmd, message
                elif message["message"].get("quick_reply"):
                    # if the response is a quick reply
                    return (
                        sender_id,
                        Cmd(message["message"]["quick_reply"].get("payload")),
                        message,
                    )
                elif message["message"].get("text"):
                    # if the response is a simple text
                    return (
                        sender_id,
                        Cmd(message["message"].get("text")),
                        message,
                    )

            if message.get("postback"):
                recipient_id = sender_id
                pst_payload = Cmd(message["postback"]["payload"])
                pst_payload.webhook = "postback"
                return recipient_id, pst_payload, message

            if message.get("read"):
                watermark = Cmd(message["read"]["watermark"])
                watermark.webhook = "read"
                return sender_id, watermark, message

            if message.get("delivery"):
                watermark = Cmd(message["delivery"]["watermark"])
                watermark.webhook = "delivery"
                return sender_id, watermark, message

            if message.get("reaction"):
                reaction = Cmd(message["reaction"]["reaction"])
                reaction.webhook = "reaction"
                return sender_id, reaction, message

            if message.get("optin"):
                optin = Cmd(message["optin"]["payload"])
                optin.webhook = "optin"
                if message["optin"].get("type") == "one_time_notif_req":
                    optin.token = message["optin"]["one_time_notif_token"]
                elif message["optin"].get("type") == "notification_messages":
                    optin.token = message["optin"]["notification_messages_token"]
                return sender_id, optin, message

    return None, Cmd(""), None


def before_run(func, **kwargs):
    res = None
    if funcs["before"] and hasattr(funcs["before"], "__call__"):
        if funcs["before"](**kwargs):
            res = func(**kwargs)
    else:
        res = func(**kwargs)

    if funcs["after"] and hasattr(funcs["after"], "__call__"):
        kwargs["res"] = res
        funcs["after"](**kwargs)

    return res


def send_next(sender_id, payload):
    chat = Messenger()
    if os.path.isfile(f"assets/private/.__{sender_id}"):
        elements = pickle.load(open(f"assets/private/.__{sender_id}", "rb"))
        if payload == "/__next":
            chat.send_generic_template(sender_id, elements[0], next=elements[1])
        else:
            chat.send_quick_reply(sender_id, elements[0], elements[1], next=elements[2])


def verif_event(testmode, payload, sender_id, message):
    if funcs["event"].get(payload.webhook):
        kw = {
            "sender_id": sender_id,
            "watermark": payload,
            "message": message,
        }
        if testmode:
            funcs["event"][payload.webhook](**kw)
        else:
            Thread(target=funcs["event"][payload.webhook], kwargs=kw).start()



import os
import json
import asyncio
import uvicorn
from .model import Model
from .logger import Logger
from threading import Thread
from .payload import Payload
from conf import Configuration  # type: ignore
from .messenger import Messenger
from fastapi.staticfiles import StaticFiles
from fastapi import FastAPI, Request, Response
from .tools import funcs, analyse, before_run, send_next, verif_event

_req = Model(init=False)
loop = asyncio.get_event_loop()

webserver = FastAPI(title="Ampalibe server")
if not os.path.isdir("assets/public"):
    os.makedirs("assets/public", exist_ok=True)

webserver.mount("/asset", StaticFiles(directory="assets/public"), name="asset")


class Init:
    def __init__(self, *args):
        self.query = Model()
        self.chat = Messenger()

    @staticmethod
    def run():
        """
        function that run framework
        """
        global _req
        _req = Model()

        Thread(target=loop.run_forever).start()

        uvicorn.run(
            "ampalibe:webserver",
            port=Configuration.APP_PORT,
            host=Configuration.APP_HOST,
        )


class Server(Request):
    """
    Content of webhook
    """

    @webserver.on_event("shutdown")
    def shutdow():
        """
        function that shutdown crontab server
        """
        loop.call_soon_threadsafe(loop.stop)

    @webserver.get("/")
    async def verif(request: Request):
        """
        Main verification for bot server is received here
        """

        if request.query_params.get("hub.verify_token") == Configuration.VERIF_TOKEN:
            return Response(content=request.query_params["hub.challenge"])
        return "Failed to verify token"

    @webserver.post("/")
    async def main(request: Request):
        """
        Main Requests for bot messenger is received here.
        """
        testmode = request.query_params.get("testmode")
        try:
            data = await request.json()
        except json.decoder.JSONDecodeError:
            return "..."

        # data analysis and decomposition
        sender_id, payload, message = analyse(data)

        if payload.webhook in ("read", "delivery", "reaction"):
            verif_event(sender_id, payload, message, testmode)
            return {"status": "ok"}

        _req._verif_user(sender_id)
        action, lang = _req.get(sender_id, "action", "lang")

        if payload in ("/__next", "/__more"):
            send_next(sender_id, payload)
            return {"status": "ok"}

        if os.path.isfile(f"assets/private/.__{sender_id}"):
            os.remove(f"assets/private/.__{sender_id}")

        payload, kw = Payload.trt_payload_in(payload)
        kw.update(
            {
                "sender_id": sender_id,
                "cmd": payload,
                "message": message,
                "lang": lang,
            }
        )

        command = funcs["command"].get(payload.split()[0])

        if command:
            _req.set_action(sender_id, None)
            if testmode:
                return before_run(command, **kw)
            else:
                Thread(
                    target=before_run,
                    args=(command,),
                    kwargs=kw,
                ).start()
        elif action:
            action, kw_tmp = Payload.trt_payload_in(action)
            kw.update(kw_tmp)

            if funcs["action"].get(action):
                if testmode:
                    return before_run(funcs["action"].get(action), **kw)
                Thread(
                    target=before_run,
                    args=(funcs["action"].get(action),),
                    kwargs=kw,
                ).start()
            else:
                Logger.error(f'⚠ Error! action "{action}" undeclared ')
        else:
            command = funcs["command"].get("/")
            if command:
                if testmode:
                    return before_run(command, **kw)
                Thread(target=before_run, args=(command,), kwargs=kw).start()
            else:
                Logger.error("⚠ Error! Default route '/' function undeclared.")
        return {"status": "ok"}
